package de.lmu.myCluster;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.text.Format;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.ArrayList;



/**
 * Outlier Detection based on local synchronization
 * Extensive Kuramoto model based synchronization
 * 
 * @author Junming Shao
 * @param None
 * Created on 16. Feb. 2010, 
 */


public class SSC{
	
    public static final double C = 1;
	private static final String ArrayList = null;
    public double T1, T2;
    public static String fname;
    public int num;
    public int dim;
    public boolean normFlag;
    public boolean noiseFlag;
    public ArrayList src; 
    int loopNum;
    int[][] biGraph;
    double[] kernelP;
    int ID;
    double[][] datan;
    double[][] data;
    int nf;
    int f;  //loop k
    double[] mdl;
    int lp;
    double minMDL;
    int nc;
    int[] countN; 
    int[][] count;
    int[] idx;
    double[]order;
    
    
    public SSC() {
    	T1 = 0.0;
    	T2 = 0.0;
    	num = 0;
    	dim = 1; 
    	normFlag = false;
    	noiseFlag = false;
    	nf = 0;
    	lp  = 0;
    	nc = 1;
    	src = new ArrayList();
    	mdl = new double[100];
    	for(int i=0;i<100;i++){
    		mdl[i] = 0.0;
    	}
    	minMDL = Double.MAX_VALUE;
    }
    /**
     * Set the related parameters
     * @param filename, the threshold for nearest neighbors of one object
     * @return none
     */  
    public void setParams(String filename, double T1, double T2){ //, boolean Flag,boolean noise){
    	fname = filename;
    	this.T1 = T1;
    	this.T2 = T2;
    	//normFlag = Flag;
    	//noiseFlag = noise;
    }
    /**
     * Load the database from a given filename
     * @param Filename
     * @return none
     */
    public void loadData(String fn){
            
        try {
        	File mFile = new File(fn);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(num);
        		dim = str.length;
        		ArrayList temp = new ArrayList(dim);    		
        		for(int i=0;i<dim;i++){
        			temp.add(Double.parseDouble(str[i]));
        		}
    			src.add(temp);
    			num = num + 1;
        	}
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        
        //Normalize the data set
        data = new double[num][dim];
        countN = new int[num];
        count = new int[num][50];
        for(int i=0;i<num;i++){
        	countN[i] = 0;
      	  for(int j=0;j<dim;j++){	  
      		data[i][j] = (Double)((ArrayList)src.get(i)).get(j); 
      	  }
         }
    }

    public void norm1(ArrayList data){
       int n = data.size();
       int d = ((ArrayList)data.get(0)).size();
       double [] max = new double[d];
       //Initialize
       for(int i=0;i<d;i++){
    	   max[i]= -1e10;
       }
       //Find the maximum value for each variance

       for(int j=0;j<d;j++){
    	  for(int i=0;i<n;i++){
    		  double temp = (Double)((ArrayList)data.get(i)).get(j);
    		  if(temp > max[j]){
    			  max[j]= temp;
    		  }    		  
    	  }
       }
       //Normalize the data set
       for(int i=0;i<n;i++){
     	  for(int j=0;j<d;j++){
     		  double temp = (Double)((ArrayList)data.get(i)).get(j);
     		  temp = temp/max[j];
     		 ((ArrayList)data.get(i)).set(j, temp);
     	  }
        }	
       //saveData(data, "normx.txt");
    }    
    
    /**
     * save the clustering result
     * @param ArrayList data (result)
     * @param String fn (filename)
     * @return none
     */
    public void norm(ArrayList data){
       int n = data.size();
       int d = ((ArrayList)data.get(0)).size();
       double [] max = new double[d];
       double [] min = new double[d];
       //Initialize
       for(int i=0;i<d;i++){
    	   max[i]= -Double.MAX_VALUE;
    	   min[i]= Double.MAX_VALUE;
       }
       //Find the maximum value for each variance

       for(int j=0;j<d;j++){
    	  for(int i=0;i<n;i++){
    		  double temp = (Double)((ArrayList)data.get(i)).get(j);
    		  if(temp > max[j]){
    			  max[j]= temp;
    		  }    		  
    		  if(temp < min[j]){
    			  min[j]= temp;
    		  } 
    	  }
       }
       //Normalize the data set
       datan = new double[n][dim];
       for(int i=0;i<n;i++){
     	  for(int j=0;j<d;j++){
     		  double temp = (Double)((ArrayList)data.get(i)).get(j);
     		  temp = (temp-min[j])/(max[j]-min[j]);     		  
     		 ((ArrayList)data.get(i)).set(j, temp);
     		datan[i][j] = temp;
     	  }
        }	
              
    }   
    
    /**
     * save the clustering result
     * @param ArrayList data (result)
     * @param String fn (filename)
     * @return none
     */
    public void saveData(ArrayList data, String fn){
        
    	int lens = data.size();
        try{
			FileOutputStream fout = new FileOutputStream(new File(fn));     
			for(int i=0;i<lens;i++){   
				ArrayList al = (ArrayList)data.get(i);
				for(int j=0; j<al.size();j++){   
					fout.write(((Double)(al.get(j))+"\t").getBytes());   
				}   
			fout.write(("\r\n").getBytes());   
			}   
        } catch (IOException ex) {
            ex.printStackTrace();
        }    		
    }

    /**
     * save the clustering result
     * @param ArrayList data (result)
     * @param String fn (filename)
     * @return none
     */
    public void saveIntData(ArrayList data, String fn){
        
    	int lens = data.size();
        try{
			FileOutputStream fout = new FileOutputStream(new File(fn));     
			for(int i=0;i<lens;i++){   
				ArrayList al = (ArrayList)data.get(i);
				for(int j=0; j<al.size();j++){   
					fout.write(((Integer)(al.get(j))+"\t").getBytes());   
				}   
			fout.write(("\r\n").getBytes());   
			}   
        } catch (IOException ex) {
            ex.printStackTrace();
        }    		
    }
    
    public void saveDoubleData(ArrayList data, String fn){
        
    	int lens = data.size();
        try{
			FileOutputStream fout = new FileOutputStream(new File(fn));     
			for(int i=0;i<lens;i++){   
				ArrayList al = (ArrayList)data.get(i);
				for(int j=0; j<al.size();j++){   
					fout.write(((Double)(al.get(j))+"\t").getBytes());   
				}   
			fout.write(("\r\n").getBytes());   
			}   
        } catch (IOException ex) {
            ex.printStackTrace();
        }    		
    }
	public void save1D(int[] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){   
			fout.write(((Integer)(data[i])+"\t").getBytes());    
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save1D(double[] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){   
			fout.write(((Double)(data[i])+"\t").getBytes());    
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save2D(double[][] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){ 
				for(int j=0;j<data[0].length;j++){
					fout.write(((Double)(data[i][j])+"\t").getBytes());  
				}
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save2D(int[][] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){ 
				for(int j=0;j<data[0].length;j++){
					fout.write(((Integer)(data[i][j])+"\t").getBytes());  
				}
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}	
    /**
     * compute the distance
     * @param double dis
     * @return none
     */    
    public double distance(double[] dis){
    	
    	double val = 0.0;
    	for(int i=0;i<dis.length;i++){
    		val = val + dis[i]*dis[i];
    	}   
    	double dist = Math.sqrt(val);
    	return dist;
    }
    
    public double[][] ArrayListtoDouble(ArrayList src){
    	int num = src.size();
    	int dim = ((ArrayList)src.get(0)).size();
    	
    	double[][] data = new double[num][dim];
    	
    	for(int i=0;i< num; i++){
    		for(int j=0;j<dim;j++){
    			data[i][j] = (Double)(((ArrayList)src.get(i)).get(j));
    		}
    	}
    	
    	return data;
    }
    
    public double LinkRangeT(ArrayList x, int d){
    	if(x.size()==0) return 0;
    	double[] temp = new double[x.size()];
    	for(int i=0;i<x.size();i++){
    		temp[i]=(Double)(((ArrayList)x.get(i)).get(d));
    	}
    	
    	double minv = Double.MAX_VALUE;
    	double maxv = -Double.MAX_VALUE;
    	
    	for(int i=0;i<x.size();i++){
    		if(temp[i]<minv) minv = temp[i];
    		if(temp[i]>maxv) maxv = temp[i];
    	}
    
    	double stepv = (maxv-minv)/x.size();
    	return stepv;
    }
    
    public void cluster(){
    	
    	int len = src.size();
    	int dimm = dim;
    	System.out.println("Number of Objects:" + len + "; Dimensionality: " + dimm);
    	boolean loop = true;
    	loopNum = 0;
    	double localOrder = 0.0;
    	double allorder = 0.0;
    	ArrayList x = src;
    	saveData(x, "synC0.txt");
    	ArrayList prex = new ArrayList();
    	
    	for(int i=0;i<len;i++){
    		ArrayList temp1 = new ArrayList(dimm);    		
    		for(int j=0;j<dimm;j++){
    			temp1.add(((ArrayList)x.get(i)).get(j));
    		}
			prex.add(temp1);
			
    	} 
    	
    	//System.out.println(((ArrayList)x.get(0)).get(0));
    	while(loop){
    		System.out.println("Loop:" + loopNum);
    		double[] order = new double[len];
    		localOrder = 0.0;
    		allorder = 0.0;
    		
    		loopNum = loopNum + 1;
    		
	    	for(int i=0;i<len;i++){	    		
	    		double[] sinValue = new double[dimm]; 
    			double[] diss = new double[dimm]; 
    			
    			double dis = 0.0;    			
	    		int num = 0; 
	    		double sita = 0.0; 
			
	    		for(int j=0;j<len;j++){	    			
	    			for(int d=0;d<dimm;d++){
	    				diss[d] = (Double)((ArrayList)(prex.get(j))).get(d)-(Double)((ArrayList)(prex.get(i))).get(d);	    				
	    			}	    			
	    			dis = distance(diss); 
	    			//System.out.println("T1=" + T1);
	    			if(dis < T1){
	    				num = num + 1;
	    				//Calculating the coupling strength    	
		    			//Weighted coupling strength of each object
	    				//kij = Math.pow(distance(temp),1);
	    				//allk = allk + kij;	
	    				//Subapce-clusters should be the objects with similar attributes, share some common things in that space
		    			for(int d=0;d<dimm;d++){
		    				if(diss[d]< T2){
		    					sinValue[d] = sinValue[d] + Math.sin(diss[d]);
		    				}
		    			}			    			
	    				sita = sita + Math.exp(-dis);
	    			}
	    		}
	    		if(num > 1){
	    			for(int d=0;d<dimm;d++){
	    				double[] x1 = new double[dimm];
	    				x1[d] = ((Double)((ArrayList)(prex.get(i))).get(d))+ ((C/num)*sinValue[d]);
		    			((ArrayList)(x.get(i))).set(d, x1[d]);    			
	    			}
	    			order[i] = sita/num;
	    		}
	    	}
	    	for(int k = 0; k < len; k++){
	    		allorder = allorder + order[k];	    		
	    	}
	    	//Local order parameter
	    	localOrder = allorder/len;
	    	
	    	saveData(x, "synC"+String.valueOf(loopNum)+".txt");
	    	
	    	
	    	//System.out.println(Double.toString(localOrder));
	    	if( localOrder > 1 - (1e-2)|| loopNum >= 50){
	    		loop = false;
	    		//System.out.println("LOCAL ORDER PARAMETER:"+Double.toString(localOrder));
	    		//System.out.println("ALL LOOPS:"+Double.toString(loopNum));
	    	}
	    	// Save the previous value of the objects of the data set
	    	
	    	for(int i=0;i<len;i++){
	    		ArrayList temp1 = new ArrayList(dimm);    		
	    		for(int j=0;j<dimm;j++){
	    			temp1.add(((ArrayList)x.get(i)).get(j));
//	    			pp[i][loopNum] = (Double)((ArrayList)x.get(i)).get(j);
	    		}
				prex.set(i, temp1);
	    	}	   	
	    	
	    	
//	    	saveData(x, "synC.txt");
	    	//this.save1D(rr, "r.txt");
//	    	this.save2D(pp, "pp.txt");
    	}
    	
    	// Save the final data
    	saveData(x, "ssc.txt"); 
    	
    	long   start1   =   System.currentTimeMillis(); 
    	NumberFormat famatter = NumberFormat.getNumberInstance();  
    	famatter.setMaximumFractionDigits(3);
    	double[][] fdata = new double[len][dim];
    	for(int i=0;i<len;i++){                 
    		for(int j=0;j<dim;j++){
	            String sData  = famatter.format((Double)((ArrayList)x.get(i)).get(j));         
	            try {
					double value = famatter.parse(sData).doubleValue();
					fdata[i][j] = value;
				} catch (ParseException e) {
					e.printStackTrace();
				} 
    		}  
    	}
    	long   end1   =   System.currentTimeMillis(); 
    	System.out.println("Format:"+Long.toString((end1-start1)/1000)+" s.");      
    	this.save2D(fdata, "fssc.txt");
	       
	    // find the distinct phase at each dimension
    	long   start2   =   System.currentTimeMillis(); 
    	
    	ArrayList al = new ArrayList();
    	ArrayList count = new ArrayList();
    	for(int d=0;d<dim;d++){
    		double phase = 0.0;
    		ArrayList temp1 = new ArrayList();
    		temp1.add(fdata[0][d]);
    		ArrayList temp2 = new ArrayList();
    		temp2.add(1);
    		for(int i=1;i<len;i++){
    			int f = 0;
    			for(int j=0;j<temp1.size();j++){
    				if(fdata[i][d]==(Double)temp1.get(j)){
    					f = 1;
    					int cur = (Integer)temp2.get(j);
    					cur = cur + 1;
    					temp2.set(j, cur);
    				}
    			}
    			if(f==0){
    				temp1.add(fdata[i][d]);
    				temp2.add(1);
    			}
    		}
    		al.add(temp1);
    		count.add(temp2);
    	}
	  
    	this.saveData(al, "phases.txt");
    	this.saveIntData(count, "count.txt");
    	
    	ArrayList selPhase = new ArrayList();
    	for(int i=0;i<count.size();i++){
    		ArrayList temp1 = (ArrayList)count.get(i);
    		ArrayList temp2 = (ArrayList)al.get(i);
    		ArrayList temp3 = new ArrayList();
    		for(int j=0;j<temp1.size();j++){
    			if((Integer)temp1.get(j)>10){
    				temp3.add((Double)temp2.get(j));
    			}
    		}
    		selPhase.add(temp3);
    	}
    	this.saveDoubleData(selPhase, "selPhase.txt");
    	
        long   end2   =   System.currentTimeMillis(); 
    	System.out.println("Find the synchronized Phases:"+Long.toString((end2-start2)/1000)+" s.");	
    	
    	//find Subspace clusters
    	findSubCluster(fdata, selPhase);
    }
    
    public void findSubCluster(double[][] fdata, ArrayList sp){
    	
    	int len = fdata.length;
    	int dim = fdata[0].length;
    	
    	for(int i=0;i<len;i++){
    		
    	}
    	
    	
    	
    }
    
        
    public double[] std(double[][] data){
    	double[] avg = new double[data[0].length];
    	double[] std = new double[data[0].length];
    	for(int i=0;i<data[0].length;i++){
    		avg[i] = 0;
    	}
    	
    	for(int j=0;j<data[0].length;j++){
    		for(int i=0;i<data.length;i++){    		
    			avg[j] = avg[j] + data[i][j];
    		}
    		avg[j] = avg[j]/data.length;
    		double s = 0.0;
    		for(int i=0;i<data.length;i++){    		
    			s = s + (data[i][j]-avg[j])*(data[i][j]-avg[j]);
    		}
    		std[j] = Math.sqrt(((s+0.0)/(data.length-1)));

    	}
    	
    	return std;
    }
    

    public double dist(ArrayList al1, ArrayList al2){
    	double res = 0.0;
    	
    	if(al1.size()!=al2.size()) return 0.0;
    	else{
    		double[] diss = new double[al1.size()];
			for(int d=0;d<al1.size();d++){
				diss[d] = (Double)(al1.get(d))-(Double)(al2.get(d));		
			}
			res = distance(diss);
    	}
    	
    	return res;
    	
    }
    public double kNN(String fn, int k){
    	
    	  double res = 0.0;
    	 

    	  
    	  // check inputs for validity
    	  if(src.size() == 0 || k<=0) return 0.0; // bail on bad input
    	 
    	  for(int s=0;s<src.size();s++){  
    		  
        	  double[] d = new double[k];
        	  int n = 0; // number of element in the res
    		  
	    	  double dd = dist((ArrayList)src.get(s),(ArrayList)src.get(0)); // load first one into list
	    	  d[n++] = dd;
	    	 
	    	  // go through all other data points
	    	  for(int i = 1; i<src.size(); i++){
	    		dd = dist((ArrayList)src.get(s),(ArrayList)src.get(i));  
	    	    if( n<k || d[k-1] > dd){ //add one
	    	      if(n<k){
	    	    	  d[n++] = dd;
	    	      }
	    	      int j = n-1;
	    	      while(j>0 && dd < d[j-1]){ // slide big data up
	    	        d[j] = d[j-1]; 
	    	        j--;
	    	      }
	    	      d[j] = dd;
	    	    }
	    	  }
	    	  
	    	  res = res + d[k-1];
	    	  
    	  }
    	  return res/src.size();
    	}

    double NN(ArrayList al){
    	double s = 0.0;
    	int n = al.size();
    	if(n<1)
    		return s;
    	
    	for(int i=0;i<n;i++){
    		double[] t1 = datan[(Integer)al.get(i)];
    		for(int j=i+1;j<n;j++){
    			double[] t2 = datan[(Integer)al.get(j)];
    			double[] t = new double[dim];
    			for(int d=0;d<dim;d++){
    				t[d] = t1[d]-t2[d];
    			}
    			s = s + distance(t);
    		}
    	}
    	
    	return s;
    	
    }
    
    
    /**
     * main function - Dynamical clustering
     */
    public static void main(String[] args) {
    	
    	long   start   =   System.currentTimeMillis(); 

    	String fn = "D1.txt"; 
    	SSC myCluster = new SSC(); 
    	myCluster.loadData(fn);
    	
    	//====Normalize the data
    	myCluster.norm(myCluster.src);
    	
    	int kk = 20;    	
    	double K0 = myCluster.kNN(fn,kk);
    	    	
    	myCluster.f = 1;
    	double K = K0;
    	myCluster.setParams(fn,K,1/10);	   
    	
    	//Find the local objects' orientation
    	
    	myCluster.cluster(); 
    	
    	
    	
  	  	long   end   =   System.currentTimeMillis(); 
  	  	System.out.println("All running time:"+Long.toString((end-start)/1000)+" s.");      
  	  	System.out.println("Ending");	
    }
}