package de.lmu.subspace.algorithm;


import java.util.ArrayList;
import java.io.Serializable;
import java.lang.Math;

import de.lmu.subspace.utils.*;


/**
 * 
 * @author Junming Shao
 * @param None
 * Created on 17. Aug. 2009, 
 */

class LocalClustering{
	public int[] id;
	public ArrayList input;
	public double T1;
	public double T2;
	public ArrayList ss;
	public ArrayList db;
	
	public LocalClustering(ArrayList data, double T1){
		this.input = data;
		this.T1 = T1;
		ss = new ArrayList();
		db = new ArrayList();
	}
	
	public int[] run(ArrayList data){
		
		int num = data.size();
		
		// calculating the pair-wise distance with all relevant dimensionality for each point
		ArrayList eachRelDim = new ArrayList();
		ArrayList<ArrayList<ArrayList<Serializable>>> allRelDim = new ArrayList();
		//ArrayList std = stdDim(data);
		ArrayList std = new ArrayList();
		for(int i=0;i<num;i++){
			for(int j=0;j<num;j++){
				if(i!=j){
					ArrayList relDim = relDimDist(((double[])data.get(i)),((double[])data.get(j)),j,std);
					if(relDim.size()>0){
						eachRelDim.add(relDim);
					}
				}
			}
			allRelDim.add(eachRelDim);
		}
		
		// find possible candidate subspace
		for(int i=0;i<num;i++){
			ArrayList cur =(ArrayList)((ArrayList)allRelDim.get(i));
			ss = subspaceSearch(cur,i);
		}
		
//		System.out.println(ss.size());
//		for(int i=0;i<ss.size();i++){
//			ArrayList al = (ArrayList)ss.get(i);
//			for(int k=0;k<al.size();k++){
//			System.out.print(" "+al.get(k)+",");
//			}
//			System.out.println("");
//		}
		
		// find the related objects in corresponding subspace
		ArrayList DB = new ArrayList();
		for(int i=0;i<ss.size();i++){
			ArrayList al = (ArrayList)ss.get(i);
			ArrayList relDB = relDBSearch(al,allRelDim);
			DB.add(relDB);
		}
		System.out.println("Totol number of objects: "+input.size());
		for(int i=0;i<DB.size();i++){
			ArrayList al = (ArrayList)DB.get(i);
			ArrayList as = (ArrayList)ss.get(i);
			System.out.print("subspace " + i +" [");
			for(int j=0;j<as.size();j++){
				System.out.print(""+as.get(j)+" ");
			}
			System.out.print("]");
			System.out.println(" "+al.size());
		}

		return id;
	}
	
	public ArrayList stdDim(ArrayList data){
		ArrayList std = new ArrayList();
		for(int i=0;i<data.size();i++){
			double stdd = 0.0;
			double mean = 0.0;
			double[] eachDim = (double[])data.get(i); 
			for(int j=0;j<eachDim.length;j++){
				mean = mean + eachDim[j];
			}
			mean = mean/eachDim.length;
			for(int j=0;j<eachDim.length;j++){
				stdd = stdd + (mean - eachDim[j])*(mean - eachDim[j]);
			}
			stdd = Math.sqrt(stdd/eachDim.length);
			std.add(stdd);
		}
		return std;
	}
	
	
    /** calculate the local relevant distance and dimensionality
     * 
     * @param double[] p,  double[] q
     */
	public ArrayList relDimDist(double[] p, double[] q, int idd, ArrayList std){
		int dim = p.length;
		double dis = 0.0;
		ArrayList neigDim = new ArrayList();
		ArrayList relDim = new ArrayList();
		//System.out.println(T1);
		for(int i=0;i<dim;i++){
			double var = Math.abs(p[i] - q[i]);
			//double stdd = ((Double) std.get(i)).doubleValue();
			//System.out.print(" " +var);
			if(var<T1){
				dis = dis + var*var;
				relDim.add(i); //the possible interesting attributes
				//System.out.print(" " + i + "");
			}
		}
		//System.out.println("");
		dis = Math.sqrt(dis);
		
		// distance invariance w.r.t. dimensionality
		dis = dis/relDim.size();
		// Put the distance,related relevant attributes respectively
		
		if(relDim.size()>0){
			neigDim.add(idd);
			neigDim.add(relDim);
			neigDim.add(dis);
			neigDim.add(relDim.size());
		}
		
		
		return neigDim;
	}
	
    /** Search all possible subspace
     * 
     * @param ArrayList ss,  all possible subspaces 
     * @param ArrayList al,  the all possible spaces of current object
     */
	public ArrayList subspaceSearch(ArrayList cur, int ID ){
		
		if(ss.size()==0){
			ArrayList temp = (ArrayList)((ArrayList)cur.get(0)).get(1);
			ss.add(temp);
			
			for(int j=1;j<cur.size();j++){
				ArrayList cl =(ArrayList)((ArrayList)cur.get(j)).get(1);
				boolean flag = true;
				for(int i=0;i<ss.size();i++){
					ArrayList al = (ArrayList)ss.get(i);
					if(cl.size()==al.size()){
						int num = 0;
						for(int k=0;k<al.size();k++){						
							if(cl.get(k)== al.get(k)){
								num = num + 1;
								flag = false;
							}
						}
						if(num==al.size()){							
							break;
						}							
					}
				}
				if(flag){ 
					ss.add(cl);
				}					
			}			
		}else
		{
			for(int j=0;j<cur.size();j++){
				ArrayList cl =(ArrayList)((ArrayList)cur.get(j)).get(1);
				boolean flag = true;
				for(int i=0;i<ss.size();i++){
					ArrayList al = (ArrayList)ss.get(i);
					if(cl.size()==al.size()){
						int num = 0;
						for(int k=0;k<al.size();k++){						
							if(cl.get(k)== al.get(k)){	
								num = num + 1;							
							}
						}
						if(num==al.size()){
							flag = false;
							break;
						}							
					}
				}
				if(flag){
					ss.add(cl);
				}					
			}
		}
		return ss;	
	}
	
    /** Find the related objects in corresponding subspace
     * 
     * @param ArrayList ss,  all possible subspaces 
     * @param ArrayList al,  the all possible spaces of current object
     */
	public ArrayList relDBSearch(ArrayList al, ArrayList allRelDim){
		
		ArrayList relDB = new ArrayList();
		int[] flagID = new int[input.size()];
		for(int nn=0;nn<flagID.length;nn++){
			flagID[nn] = 0;
		}
		for(int i=0;i<allRelDim.size();i++){
			ArrayList eachObjectDim =(ArrayList) allRelDim.get(i);
			
			for(int j=0;j<eachObjectDim.size();j++){
				int idd = (Integer)((ArrayList)eachObjectDim.get(j)).get(0);
				if(flagID[idd]==0){
					ArrayList ll = (ArrayList)((ArrayList)eachObjectDim.get(j)).get(1);
					
					int num = 0;
					if(ll.size()==al.size()){
						for(int n=0;n<ll.size();n++){
							if(ll.get(n)== al.get(n)){	
								num = num + 1;							
							}
						}
						if(num==al.size()){
							relDB.add(((ArrayList)eachObjectDim.get(j)).get(0));
							flagID[idd] = 1;
						}	
					}
					if(ll.size()>al.size()){
						for(int m=0;m<al.size();m++){
							for(int p=0;p<ll.size();p++){
								if(ll.get(p)== al.get(m)){
									num = num + 1;
									flagID[idd] = 1;
								}
							}
							if(num==al.size()){
								relDB.add(((ArrayList)eachObjectDim.get(j)).get(0));
							}						
						}
					}
				}		
			}
		}
	
		return relDB;	
	}
	
    public static void main(String arg[]){
    	
		String fn = "D1.txt";
    	System.out.println("Reading data from TXT filename...");
    	ReadData res = new ReadData();
    	ArrayList data = res.readTxt(fn);
    	
    	long   start   =   System.currentTimeMillis(); 
    	int[] id = new int[data.size()];
    	LocalClustering lc = new LocalClustering(data, 0.2);
    	id = lc.run(data);
    	
		long   end   =   System.currentTimeMillis(); 		
	  	System.out.println("All running time:"+Long.toString(end-start)+" ms.");
	    System.out.println("Ending");	    

    }
}