package de.lmu.subspace.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
/**
 * Read data
 * 
 * @author Junming Shao
 * @param None
 * Created on 17. Mar. 2009, 
 */

public class ReadData{
	public int[] data;
	
	public ReadData(){
		
	}

    /**
     * Load the database from a given filename
     * @param Filename
     * @return none
     */
    public int[] readClusterID(String fn,int num){
    	
    	int row = 0; 
    	int[] data = new int[num];
        try {
        	File mFile = new File(fn);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(str[1]);
        		data[row]= Integer.parseInt(str[0]);
        		row = row + 1;
        	}
        } catch (IOException ex) {
            ex.printStackTrace();
        } 
        return data;
    }
    
    /**
     * Load the .txt file from a given filename
     * @param Filename
     * @return none
     */
    public ArrayList readTxt(String fn){
    	
    	int num = 0; 
    	ArrayList data = new ArrayList();
        try {
        	File mFile = new File(fn);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(str[1]);
        		double[] temp = new double[str.length];
        		for(int k=0;k<str.length;k++){
        			temp[k]= Double.parseDouble(str[k]);
        		}
        		data.add(temp);
        	}
        } catch (IOException ex) {
            ex.printStackTrace();
        } 
        return data;
    }
    
    /**
     * Load the 3-D points from a given filename
     * @param Filename
     * @return none
     */
    public float[] readVTKPoint(String fn,int num){
    	
    	int row = 0; 
    	float[] data = new float[num];
        try {
        	File mFile = new File(fn);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(str[1]);
        		for(int k=0;k<str.length;k++){
        			data[row]= Float.parseFloat(str[k]);
        			row = row + 1;
        		}
        		
        	}
        	if(row==num){
        		System.out.println("Reading Points Sucessfully");
        	}
        } catch (IOException ex) {
            ex.printStackTrace();
        } 
        return data;
    }
    /**
     * Load the 3-D fibers from a given filename
     * @param Filename
     * @return none
     */
    public ArrayList readVTK(String fp, String fl, int num){
    	
    	int row = 0; 
    	float[] data = new float[num];
    	ArrayList fibers = new ArrayList();
        try {
        	File mFile = new File(fp);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(str[1]);
        		for(int k=0;k<str.length;k++){
        			data[row]= Float.parseFloat(str[k]);
        			row = row + 1;
        		}
        		
        	}
        	if(row==num){
        		System.out.println("Reading Points Sucessfully");
        		System.out.println("Number:"+row);
        	}
        	
        	//Reading Fibers
        	File lFile = new File(fl);
        	FileReader fline = new FileReader(lFile);
        	BufferedReader brline = new BufferedReader(fline);
        	String line2;
        	int numFiber = 0;
        	while((line2=brline.readLine())!=null){
        		line2 = line2.trim();        		
        		String[] str = line2.split("\\s+");
        		//System.out.println(str[1]);
        		int len = Integer.parseInt(str[0]);
        		int start = Integer.parseInt(str[1]);
        		start = start*3;
        		double[][] fiber = new double[len][3];
        		for(int i=0;i<len;i++){        			
        			fiber[i][0]= data[start];
        			fiber[i][1]= data[start+1];
        			fiber[i][2]= data[start+2];
        			start = start + 3;
        		}     		
        		fibers.add(fiber); 
        		numFiber = numFiber + 1;
        	}
        	if(numFiber==841){
        		System.out.println("Reading Fibers Sucessfully");
        		System.out.println("Number of Fibers:"+numFiber);
        	}
        	
        } catch (IOException ex) {
            ex.printStackTrace();
        } 
        return fibers;
    }
    
    public static void main(String arg[]){
    	
		String fp = "pointSet3.txt";
		String fl = "lineSet3.txt";
    	System.out.println("Reading data from VTK format...");
    	ReadData res = new ReadData();
    	ArrayList data = res.readVTK(fp,fl,86731*3);
    	double[][] temp = (double[][])(data.get(500));
    	System.out.println(temp.length);
    	for(int i=0;i<((double[][])(data.get(500))).length;i++){
    		
    		System.out.println(" "+ temp[i][0]+" "+ temp[i][1]+" "+temp[i][2]+" ");
    	}
    }
}